# SIS with a fixed recovery time
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

class SIS_FixedRecovery(SIS):
    '''The Susceptible-Infected-Susceptible :term:`compartmented model of disease`,
    in the variation where the time spent infected is fixed rather than happening
    with some probability.'''
    
    # the additional model parameter
    T_INFECTED = 'tInfected'           #: Parameter for the time spent infected before becoming susceptible again.

    # node attribute for infection time
    INFECTION_TIME = 'infection_time'  #: Attribute recording when a node became infected
    
    def __init__( self ):
        super(SIS_FixedRecovery, self).__init__()

    def build( self, params ):
        '''Build the variant SIS model. The difference between this and the
        reference :class:`SIS` model is that only infection events happen
        probabilistically, with recovery events happening on a fixed schedule
        depending on the :attr:`T_INFECTED` parameter.

        :param params: the model parameters'''
        pInfected = params[self.P_INFECTED]
        pInfect = params[self.P_INFECT]
        self._tInfected = params[self.T_INFECTED]
        
        self.addCompartment(self.SUSCEPTIBLE, 1 - pInfected)
        self.addCompartment(self.INFECTED, pInfected)

        self.addLocus(self.SUSCEPTIBLE, self.INFECTED, name = self.SI)
        self.addEvent(self.SI, pInfect, lambda t, g, e: self.infect(t, g, e))

    def setUp( self, dyn, g, params ):
        '''After setting up as normal, post recovery events for any nodes that are
        initially infected.

        :param dyn: the dynamics
        :param g: the network
        :param params: the simulation parameters'''
        super(SIS_FixedRecovery, self).setUp(dyn, g, params)

        # traverse the set of initially-infected nodes
        tInfected = params[self.T_INFECTED]
        for n in self.compartment(g, self.INFECTED):
            # record that the node was initially infected
            g.node[n][self.INFECTION_TIME] = 0.0
        
            # post the recovery event
            dyn.postEvent(tInfected, g, n, lambda d, t, g, e: self.recover(d, t, g, e))          

    def infect( self, dyn, t, g, (n, m) ):
        '''Perform the normal infection event, and then post an event to recover
        the infected node back to susceptible at the appropriate time.

        :param dyn: the dynamics
        :param t: the simulation time
        :param g: the network
        :param e: the edge transmitting the infection, susceptible-infected'''

        # infect as normal
        super(SIS_FixedRecovery, self).infect(dyn, t, g, (n, m))

        # record the infection time
        g.node[n][self.INFECTION_TIME] = t
        
        # post the recovery event for the appropriate time in the future
        dyn.postEvent(t + self._tInfected, g, n, lambda d, t, g, e: self.recover(d, t, g, e))
                
                
                
   



# Test compartmented model basic functions
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

import epyc
import unittest
import networkx
import traceback

class CompartmentedModelTest(unittest.TestCase):

    def setUp( self ):
        '''Set up the experimental parameters and experiment.'''
        self._er = networkx.erdos_renyi_graph(1000, 0.005)
        self._params = dict(pInfect = 0.1,
                            pInfected = 0.01,
                            pRemove = 0.05)
       
    def testPopulation( self ):
        '''Test populating a model.'''
        m = SIR()
        e = CompartmentedStochasticDynamics(m, self._er)
        e.setUp(self._params)
                
    def testLoci( self ):
        '''Test we can populate loci correctly.'''
        m = SIR()
        g = networkx.Graph()
        g.add_edges_from([ (1, 2), (2, 3), (1, 4), (3, 4) ])
        e = CompartmentedStochasticDynamics(m, g)
        params = dict(pInfect = 0.1,
                      pInfected = 1.0,    # infect all nodes initially
                      pRemove = 0.05)
        e.setUp(params)

        # keep track of the other compartments as well
        e._model.addLocus(SIR.SUSCEPTIBLE)
        e._model.addLocus(SIR.REMOVED)
        
        # all nodes in I
        self.assertItemsEqual(e._model._loci[SIR.INFECTED].elements(), [ 1, 2, 3, 4 ])

        # one node from I into S, two edges into SI
        m.changeCompartment(e.network(), 1, SIR.SUSCEPTIBLE)
        self.assertItemsEqual(e._model._loci[SIR.INFECTED].elements(), [ 2, 3, 4 ])
        self.assertItemsEqual(e._model._loci[SIR.SUSCEPTIBLE].elements(), [ 1 ])
        self.assertItemsEqual(e._model._loci[SIR.SI].elements(), [ (1, 2), (1, 4) ])

        # recover the infected node
        m.changeCompartment(e.network(), 1, SIR.REMOVED)
        self.assertItemsEqual(e._model._loci[SIR.INFECTED].elements(), [ 2, 3, 4 ])
        self.assertItemsEqual(e._model._loci[SIR.SUSCEPTIBLE].elements(), [])
        self.assertItemsEqual(e._model._loci[SIR.REMOVED].elements(), [ 1 ])
        self.assertItemsEqual(e._model._loci[SIR.SI].elements(), [ ])        



# Test the SIR and SIS fixed-time recovery variants
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *
from .compartmenteddynamics import CompartmentedDynamicsTest

import epyc
import unittest
import networkx
import traceback

class FixedRecoveryTest(unittest.TestCase):

    def setUp( self ):
        '''Set up the experimental parameters and experiment.'''
        
        # single experiment
        self._params = dict(pInfect = 0.1,
                            pInfected = 0.01,
                            tInfected = 1)
        self._network = networkx.erdos_renyi_graph(1000, 0.005)

        # lab run
        self._lab = epyc.Lab()
        self._lab['pInfect'] = [ 0.1, 0.2, 0.3 ]
        self._lab['pInfected'] = [ 0.01 ]
        self._lab['tInfected'] = [ 0.5, 1, 2 ]

    def testRunSingleSIRSynchronous( self ):
        '''Test a single run of a fixed-period SIR under synchronous dynamics.'''
        e = CompartmentedSynchronousDynamics(SIR_FixedRecovery(), self._network)
        rc = e.set(self._params).run()
        if not rc[epyc.Experiment.METADATA][epyc.Experiment.STATUS]:
            print rc[epyc.Experiment.METADATA][epyc.Experiment.EXCEPTION]
            traceback.print_tb(rc[epyc.Experiment.METADATA][epyc.Experiment.TRACEBACK])

    def testRunSingleSIRStochastic( self ):
        '''Test a single run of a fixed-period SIR under stochastic dynamics.'''
        e = CompartmentedStochasticDynamics(SIR_FixedRecovery(), self._network)
        rc = e.set(self._params).run()
        if not rc[epyc.Experiment.METADATA][epyc.Experiment.STATUS]:
            print rc[epyc.Experiment.METADATA][epyc.Experiment.EXCEPTION]
            traceback.print_tb(rc[epyc.Experiment.METADATA][epyc.Experiment.TRACEBACK])
  



# Base class for testing compartmented models under different dynamics
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

import epyc
import unittest
import networkx
import traceback

class CompartmentedDynamicsTest(unittest.TestCase):

    def setUp( self ):
        '''Set upo for the specific model under test. This should fill in
        self._network, self._model, self._params, self._maxTime, 
        and self._lab as required.'''
        raise NotYetImplementyed('setUp')

    def testRunSingleStochastic( self ):
        '''Test a single run of a stochastic dynamics.'''
        e = CompartmentedStochasticDynamics(self._model, self._network)
        if self._maxTime is not None:
            e.setMaximumTime(self._maxTime)
        rc = e.set(self._params).run()
        if not rc[epyc.Experiment.METADATA][epyc.Experiment.STATUS]:
            print rc[epyc.Experiment.METADATA][epyc.Experiment.EXCEPTION]
            traceback.print_tb(rc[epyc.Experiment.METADATA][epyc.Experiment.TRACEBACK])
    
    def testRunSingleSynchronous( self ):
        '''Test a single run of a synchronous dynamics.'''
        e = CompartmentedSynchronousDynamics(self._model, self._network)
        if self._maxTime is not None:
            e.setMaximumTime(self._maxTime)
        rc = e.set(self._params).run()
        if not rc[epyc.Experiment.METADATA][epyc.Experiment.STATUS]:
            print rc[epyc.Experiment.METADATA][epyc.Experiment.EXCEPTION]
            traceback.print_tb(rc[epyc.Experiment.METADATA][epyc.Experiment.TRACEBACK])
        
    @unittest.skip('not yet')
    def testRunMultipleStochastic( self ):
        '''Test a stochastic dynamics run of a model over a (small) parameter space.'''
        e = CompartmentedStochasticDynamics(self._model, self._network)
        if self._maxTime is not None:
            e.setMaximumTime(self._maxTime)
        self._lab.runExperiment(e)

    @unittest.skip('not yet')
    def testRunMultipleSynchronous( self ):
        '''Test a synchronousdynamics run of a model over a (small) parameter space.'''
        e = CompartmentedSynchronousDynamics(self._model, self._network)
        if self._maxTime is not None:
            e.setMaximumTime(self._maxTime)
        self._lab.runExperiment(e)



# Networks dynamics simulation base class
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import epyc
import networkx
from heapq import *

class Dynamics(epyc.Experiment, object):
    '''A dynamical process over a network. This is the abstract base class
    for implementing different kinds of dynamics as computational experiments
    suitable for running under. Sub-classes provide synchronous and stochastic
    (Gillespie) simulation dynamics.'''

    # Additional metadata elements
    TIME = 'simulation_time'      #: Metadata element holding the logical simulation end-time.
    EVENTS = 'simulation_events'  #: Metadata element holding the number of events that happened.

    # the default maximum simulation time
    DEFAULT_MAX_TIME = 20000      #: Default maximum simulation time.
    
    def __init__( self, g = None ):
        '''Create a dynamics, optionally initialised to run on the given network.
        The network (if provided) is treated as a prototype that is copied before
        each individual simulation experiment.
        
        :param g: prototype network (optional)'''
        super(Dynamics, self).__init__()
        self._graphPrototype = g
        self._graph = None
        self._maxTime = self.DEFAULT_MAX_TIME
        self._posted = []

    def network( self ):
        '''Return the network this dynamics is running over.

        :returns: the network'''
        return self._graph

    def setNetworkPrototype( self, g ):
        '''Set the network the dynamics will run over. This will be
        copied for each run of an individual experiment.

        :param g: the network'''
        self._graphPrototype = g

    def setMaximumTime( self, t ):
        '''Set the maximum default simulation time. The default is given
        by :attr:`DEFAULT_MAX_TIME`.

        param: t: the maximum time'''
        self._maxTime = t
        
    def at_equilibrium( self, t ):
        '''Test whether the model is an equilibrium. Override this method to provide
        alternative and/or faster simulations.
        
        :param t: the current simulation timestep
        :returns: True if we're done'''
        return (t >= self._maxTime)

    def setUp( self, params ): 
        '''Before each experiment, create a working copy of the prototype network.

        :param params: parameters of the experiment'''

        # perform the default setup
        super(Dynamics, self).setUp(params)

        # make a copy of the network prototype
        self._graph = self._graphPrototype.copy()

        # empty the queue of posted events
        self._posted = []

    def tearDown( self ):
        '''At the end of each experiment, throw away the copy.'''

        # perform the default tear-down
        super(Dynamics, self).tearDown()

        # throw away the worked-on model
        self._graph = None

    def eventDistribution( self, t ):
        '''Return the event distribution, a sequence of (l, p, f) triples
        where l is the :term:`locus` of the event, p is the probability of an
        event occurring, and f is the :term:`event function` called to make it
        happen. This method must be overridden in sub-classes.
        
        It is perfectly fine for an event to have a zero probability.

        :param t: current time
        :returns: the event distribution'''
        raise NotYetImplemented('eventDistribution()')

    def postEvent( self, t, g, e, ef ):
        '''Post an event to happen at time t. The :term:`event function` should
        take the simulation time, network, and element for the event. At time t
        it is called with the given network and element.

        :param t: the current time
        :param g: the network
        :param e: the element (node or edge) on which the event occurs
        :param ef: the event function'''
        heappush(self._posted, (t, (lambda: ef(self, t, g, e))))

    def _nextPendingEventBefore( self, t ):
        '''Return the next pending event to occur at or before time t.

        :param t: the current time
        :returns: a pending event function or None'''
        if len(self._posted) > 0:
            # we have events, grab the soonest
            (et, pef) = heappop(self._posted)
            if et <= t:
                # event should have occurred, return it
                return pef
            else:
                # this (and therefore all further events) are in the future, put it back
                heappush(self._posted, (et, pef))
                return None
        else:
            # we don't have any events
            return None
        
    def pendingEvents( self, t ):
        '''Retrieve any :term:`posted event` pending to be executed at or
        before time t. The pending events are returned in the form of
        zero-argument functions that can simply be called to execute
        the events. The events are returned as a list, with the
        earliest-posted event first.

        Be aware that running the returned events in order may not be enough to
        accurately run the simulation in the case where firing an
        event causes another event to be posted before t. It may be
        easier to use :meth:`runPendingEvents` to run all pending
        events, which handles this case automatically.

        :param t: the current time
        :returns: a (possibly empty) list of pending event functions'''
        pending = []
        while True:
            pef = self._nextPendingEventBefore(t)
            if pef is None:
                # no more events pending, return those we've got
                return pending
            else:
                # store the pending event function
                pending.append(pef)

    def runPendingEvents( self, t):
        '''Retrieve and fire any pending events at time t. This method handles
        the case where firing an event posts another event that needs to be run
        before other already-posted events coming before time t: in other words,
        it ensures that the simulation order is respected.

        :param t: the current time
        :returns: the number of events fired'''
        n = 0
        while True:
            pef = self._nextPendingEventBefore(t)
            if pef is None:
                # no more pending events, return however many we've fired already
                return n
            else:
                # fire the event
                pef()
                n = n + 1
                



# Synchronous dynamics base class
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

import epyc
import networkx
import numpy
from copy import copy

class SynchronousDynamics(Dynamics):
    '''A dynamics that runs synchronously in discrete time, applying local
    rules to each node in the network. These are simple to understand and
    simple to code for many cases, but can be statistically inexact and slow
    for large systems.'''

    # additional metadata
    TIMESTEPS_WITH_EVENTS = 'timesteps_with_events'  #: Metadata element holding the number timesteps that actually had events occur within them

    def __init__( self, g = None ):
        '''Create a dynamics, optionally initialised to run on the given prototype
        network.
        
        :param g: prototype network to run over (optional)'''
        super(SynchronousDynamics, self).__init__(g)

    def do( self, params ):
        '''Synchronous dynamics. We apply :meth:`dynamics` at each timestep
        and then check for completion using :meth:`at_equilibrium`.
        
        :param params: the parameters of the simulation
        :returns: a dict of experimental results'''
        
        # run the dynamics
        g = self.network()
        t = 0
        events = 0
        timestepEvents = 0
        while not self.at_equilibrium(t):
            # fire any events posted for at or before this time
            nev = self.runPendingEvents(t)
            
            # retrieve all the events, their loci, probabilities, and event functions
            dist = self.eventDistribution(t)

            # run through all the events in the distribution
            for (l, p, ef) in dist:
                if p > 0.0:
                    # run through every possible element on which this event may occur
                    for e in copy(l.elements()):
                        # test for occurrance of the event on this element
                        if numpy.random.random() <= p:
                            # yes, perform the event
                            ef(self, t, g, e)
                            
                            # update the event count
                            nev = nev + 1

            # add the events to the count
            events = events + nev
            if nev > 0:
                # we had events happen in this timestep
                timestepEvents = timestepEvents + 1

            # advance to the next timestep
            t = t + 1

        # run any events posted for before the maximum simulation time
        self.runPendingEvents(self._maxTime)

        # add some more metadata
        (self.metadata())[self.TIME] = t
        (self.metadata())[self.EVENTS] = events
        (self.metadata())[self.TIMESTEPS_WITH_EVENTS] = timestepEvents

        # report results
        rc = self.experimentalResults()
        return rc



# Initialisation for epydemic
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

'''`epydemic` is a library for performing simulations for a range of
epidemic spreeading (and other) processes, simulated over networks
represented using `networkx`.

Epidemic processes are very important in both network science and its
applications. The most common application is to study the was in which
diseases progress in different network conditions, depending on their
infectiousness and other properties. Typically such processes are
modelled as a :term:`compartmented model of disease` with conditional
probabilities for moving between compartments (familiar to
computer scientists as stochastic finite state machines).

`epydemic` provides the basic simulation machinery for performing
epidemic simulations under two different simulation regimes:
synchronous :term:`discrete time` simulation in which time proceeds in
discrete time intervals, and stochastic or Gillespie :term:`continuous
time` simulations which are better for handling a wider range of
:term:`event` probabilities (but which are slightly harder to
specify).

'''
 
# networks with dynamical processes
from .networkdynamics import Dynamics
from .synchronousdynamics import SynchronousDynamics
from .stochasticdynamics import StochasticDynamics

# compartmented models
from .compartmentedmodel import CompartmentedModel
from .loci import Locus, NodeLocus, EdgeLocus
from .compartmentedsynchronousdynamics import CompartmentedSynchronousDynamics
from .compartmentedstochasticdynamics import CompartmentedStochasticDynamics

# reference disease models
from .sir_model import SIR
from .sis_model import SIS

# variant disease models
from .sir_model_fixed_recovery import SIR_FixedRecovery
from .sis_model_fixed_recovery import SIS_FixedRecovery



# Test network dynamics
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epydemic import *

import epyc
import unittest
import networkx

class NetworkDynamicsTest(unittest.TestCase):

    def testPosting( self ):
        '''Test posting events.'''
        dyn = Dynamics()

        # event function builder, will increment the shared
        # value when fired
        self._v = 0
        def make_ef( w ):
            def ef( dyn, t, g, e ):
                self._v = self._v + w
            return ef
        
        # post some events at different times
        dyn.postEvent(1, None, None, make_ef(1))
        dyn.postEvent(2, None, None, make_ef(20))
        dyn.postEvent(3, None, None, make_ef(100))
        dyn.postEvent(4, None, None, make_ef(200))

        # check no events before the first one posted
        self.assertItemsEqual(dyn.pendingEvents(0.5), [])

        # check firing of earliest event
        pefs = dyn.pendingEvents(1)
        self.assertTrue(len(pefs), 1)
        (pefs[0])()                # fire the event
        self.assertTrue(self._v, 1)
        self.assertItemsEqual(dyn.pendingEvents(1), [])

        # check multiple events coming off in the right order
        pefs = dyn.pendingEvents(3)
        self.assertTrue(len(pefs), 2)
        (pefs[0])()
        self.assertTrue(self._v, 21)
        (pefs[1])()
        self.assertTrue(self._v, 121)
        self.assertItemsEqual(dyn.pendingEvents(3), [])

        # check we can run all remaining events
        dyn.runPendingEvents(10)       # a long time in the future
        self.assertTrue(self._v, 321)
        self.assertItemsEqual(dyn.pendingEvents(10), [])
        self.assertItemsEqual(dyn.pendingEvents(20), [])
        
    def testPostedPosting( self ):
        '''Test the case when a posted event itself posts an event.'''
        dyn = Dynamics()

        # event function builder, will increment the shared
        # value when fired and also post a further event
        self._v = 0
        def make_ef( dyn, dt ):
            def ef( dyn, t, g, e ):
                self._v = self._v + 1
                dyn.postEvent(t + dt, None, None, make_ef(dyn, dt))
            return ef
        
        # post the first of the events
        dyn.postEvent(1, None, None, make_ef(dyn, 5))

        # post an intermediate event that should be fired alongside a regular one
        def inter( dyn, t, g, e ):
            self._v = self._v + 1000
        dyn.postEvent(12, None, None, inter)

        # carefully run the event sequence for a while
        for t in range(1, 25, 5):
            nev = dyn.runPendingEvents(t)
            v = self._v
            if t == 1:
                self.assertEqual(v, 1)
                self.assertEqual(nev, 1)
            if t == 6:
                self.assertEqual(v, 2)
                self.assertEqual(nev, 1)
            if t == 11:
                self.assertEqual(v, 3)
                self.assertEqual(nev, 1)
            if t == 16:
                self.assertEqual(v, 1004)
                self.assertEqual(nev, 2)
            if t == 21:
                self.assertEqual(v, 1005)
                self.assertEqual(nev, 1)

        # should be one event left
        self.assertEqual(len(dyn.pendingEvents(100)), 1)
        



# Compartmented models model
#
# Copyright (C) 2017 Simon Dobson
# 
# This file is part of epydemic, epidemic network simulations in Python.
#
# epydemic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epydemic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.

import epydemic
from epydemic import *

import numpy
import random
import collections

class CompartmentedModel(object):
    '''The base class for compartmented models. A :term:`compartmented model of disease`
    represents a disease as a collection of discrete :term:`compartments`, with each 
    compartment specifying some facet of the disease's progression. Nodes
    transition between compartments with some probability, so the disease's
    progession is a stochastic process in which the various nodes typically move
    through several compartments according to the probabilities defined.'''
    
    # model state variables
    COMPARTMENT = 'dynamics_compartment'   #: Node attribute used to hold compartment.
    OCCUPIED = 'occupied'                  #: Edge attribute, True if edge passed infection.
    
    def __init__( self ):
        '''Create a new compartmented model.'''
        super(CompartmentedModel, self).__init__()
        self._compartments = []
        self._loci = dict()
        self._effects = dict()
        self._events = []
        
    def addCompartment( self, c, p = 0.0 ):
        '''Add a compartment to the model. A node is assigned to the compartment
        initially with the given probability. The probabilities for all compartments
        in the model must sum to 1.

        :param c: the compartment name
        :param p: the initial occupancy probability (defaults to  0.0)'''
        self._compartments.append((c, p))

    def addLocus( self, l, r = None, name = None ):
        '''Add a one- or two-compartment locus, corresponding to a node
        or edge at which dynamics occurs (e.g,. which can have events associated
        with it).

        :param l: the left-hand (or only) compartment
        :param r: the right-hand compartment (for an edge)
        :param name: the name of the locus'''

        if r is None:
            # one compartment, node locus
            if name is None:
                name = l
            if name in self._loci.keys():
                raise Exception('Locus {n} already exists'.format(n = name))

            # add locus
            locus = epydemic.NodeLocus(name, l)
            self._loci[name] = locus

            # add effects
            self._addHandlers(l, lambda m, g, n, c: locus.leaveHandler(m, g, n, c),
                                 lambda m, g, n, c: locus.enterHandler(m, g, n, c))
        else:
            # two compartments, edge locus
            if name is None:
                name = '{l}{r}'.format(l = l, r = r)
                if name in self._loci.keys():
                    raise Exception('Locus {n} already exists'.format(n = name))

            # add locus
            locus = epydemic.EdgeLocus(name, l, r)
            self._loci[name] = locus

            # add effects
            self._addHandlers(l, lambda m, g, n, c: locus.leaveHandler(m, g, n, c),
                                 lambda m, g, n, c: locus.enterHandler(m, g, n, c))
            self._addHandlers(r, lambda m, g, n, c: locus.leaveHandler(m, g, n, c),
                                 lambda m, g, n, c: locus.enterHandler(m, g, n, c))

    def _addHandlers( self, c, lh, eh ):
        '''Add handler functions for a compartment, to be called whenever the contents
        of that compartment change. This links compartments to loci.

        Handlers are functions that take a model, a network, a node, and a
        compartment, and provide the appropriate behaviour for when the node
        enters (or leaves) the compartment.
        
        :param c: the compartment
        :param lh: the leave handler
        :param eh: the enter handler'''
        if c not in self._effects.keys():
            self._effects[c] = [ (lh, eh) ]
        else:
            self._effects[c].append((lh, eh))

    def _callLeaveHandlers( self, g, n, c ):
        '''Call all handlers affected by a node leaving a compartment.

        :param g: the network
        :param n: the node
        :param c: the compartment'''
        if c in self._effects.keys():
            for (lh, _) in self._effects[c]:
                lh(self, g, n, c)

    def _callEnterHandlers( self, g, n, c ):
        '''Call all handlers affected by a node entering a compartment.

        :param g: the network
        :param n: the node
        :param c: the compartment'''
        if c in self._effects.keys():
            for (_, eh) in self._effects[c]:
                eh(self, g, n, c)

    def compartment( self, g, c ):
        '''Return all the nodes currently in a particular compartment in a network.

        :param g: the network
        :param c: the compartment
        :returns: a collection of nodes'''
        return [ n for n in g.nodes_iter() if g.node[n][self.COMPARTMENT] == c]
    
    def changeCompartment( self, g, n, c ):
        '''Change the compartment of a node. This will update all loci
        potentiually affected by the change.

        :param g: the network
        :param n: the node
        :param c: the new compartment for the node'''
        
        # propagate effects of leaving the current compartment
        self._callLeaveHandlers(g, n, g.node[n][self.COMPARTMENT])

        # record new compartment on node
        g.node[n][self.COMPARTMENT] = c

        # propagate effects of entering new compartment
        self._callEnterHandlers(g, n, c)

    def addEvent( self, l, p, ef ):
        '''Add a probabilistic event at a locus, occurring with a particular (fixed)
        probability and calling the :term:`event function` when it is selected.

        The :term:`event function` takes the dynamics, the simulation time,
        the network, and the element to which the event applies (a node or an edge,
        which will have been selected from the event's locus).

        :param l: the locus name
        :param p: the event probability
        :param ef: the event function'''
        self._events.append((self._loci[l], p, ef))

    def initialCompartmentDistribution( self ):
        '''Return the initial distribution of nodes to compartments. The
        result should be a valid distribution, with probabilities summing
        to one.

        :returns: a list of (compartment, probability) pairs'''
        return self._compartments

    def eventDistribution( self, t ):
        '''Return the distribution of events. The result should be a valid distribution,
        with probabilities summing to one. The default implemnentation returns a constant
        distribution populated from the model parameters. Sub-classes may override to,
        for example, provide time-evolving probabilities.

        :param t: the current simulation time
        :returns: a list of (locus, probability, event function) triples'''
        return self._events

    def setUp( self, dyn, g, params ):
        '''Set up the initial population of nodes into compartments.
        
        :param dyn: the dynamics
        :param g: the network
        :param params: the simulation parameters'''

        # initialise all nodes to an empty compartment
        # (so we can assume all nodes have a compartment attribute)
        for n in g.nodes_iter():
            g.node[n][self.COMPARTMENT] = None
            
        # get the initial compartment distribution
        dist = self.initialCompartmentDistribution()

        # assign nodes to compartments
        for n in g.nodes_iter():
            # select a compartment according to the initial distribution
            r = numpy.random.random()
            a = 0.0
            for (c, p) in dist:
                a = a + p
                if r <= a:
                    # change node's compartment
                    self.changeCompartment(g, n, c)
                    
                    # on to the next node
                    break

        # mark edges as unoccupied
        for (_, _, data) in g.edges_iter(data = True):
            data[self.OCCUPIED] = False

    def markOccupied( self, g, e ):
        '''Mark the given edge as having been occupied by the dynamics, i.e., to
        have been traversed in transmitting the disease.

        :param g: the network
        :param e: the edge'''
        (n, m) = e
        data = g.get_edge_data(n, m)
        data[self.OCCUPIED] = True
        
    def results( self, g ):
        '''Create a dict of experimental results for the experiment.

        :param g: the network
        :returns: a dict of experimental results'''
        rc = dict()

        # add final sizes of all loci
        rc['loci'] = dict()
        for (l, locus) in self._loci.iteritems():
            rc['loci'][l] = len(locus)

        # add final sizes of all compartments
        rc['compartments'] = dict()
        for (c, _) in self._compartments:
            rc['compartments'][c] = 0        
        for n in g.nodes_iter():
            c = g.node[n][self.COMPARTMENT]
            rc['compartments'][c] = rc['compartments'][c] + 1

        return rc

    def skeletonise( self, g ):
        '''Remove unoccupied edges from the network. This leaves the network
        consisting of only "occupied" edges that were used to transmit the
        infection between nodes. Note that this process means that further
        dynamics over the network doesn't make sense.
        
        :returns: the network with unoccupied edges removed'''
        
        # find all unoccupied edges
        edges = []
        for (n, m, data) in g.edges_iter(data = True):
            if (self.OCCUPIED not in data.keys()) or (data[self.OCCUPIED] != True):
                # edge is unoccupied, mark it to be removed
                # (safe because there are no parallel edges)
                edges.insert(0, (n, m))
                    
        # remove all the unoccupied edges
        g.remove_edges_from(edges)
        
        return g
    
